@@ -5,10 +5,9 @@ Represent data processing expressions.
 import abc
 from typing import Any, List, Optional, Set, Union
 
-import numpy
-
 import data_algebra.util
-import data_algebra.pandas_model
+from data_algebra.expression_walker import ExpressionWalker
+import data_algebra.data_model
 
 
 # for some ideas in capturing expressions in Python see:
@@ -143,20 +142,6 @@ class PreTerm(abc.ABC):
         """
         # can't use == as that builds a larger expression
 
-    # tree re-write
-
-    @abc.abstractmethod
-    def get_views(self):
-        """
-        return list of unique views, expectation list is of size zero or one
-        """
-
-    @abc.abstractmethod
-    def replace_view(self, view):
-        """
-        Move expression to a new view.
-        """
-
     # analysis
 
     def get_column_names(self, columns_seen: Set[str]) -> None:
@@ -178,9 +163,9 @@ class PreTerm(abc.ABC):
     # eval
 
     @abc.abstractmethod
-    def evaluate(self, data_frame):
+    def act_on(self, arg, *, expr_walker: ExpressionWalker):
         """
-        Evaluate expression, taking data from data_frame.
+        Apply expression to argument.
         """
 
     # emitters
@@ -789,7 +774,7 @@ class Term(PreTerm, abc.ABC):
 
     def is_bad(self):
         """
-        Return which items are bad (null or nan) (vectorized).
+        Return which items in a numeric column are bad (null, None, nan, or infinite) (vectorized).
         """
         return self.__uop_expr__("is_bad")
 
@@ -1060,24 +1045,12 @@ class Value(Term):
             return False
         return self.value == other.value
 
-    def get_views(self):
+    def act_on(self, arg, *, expr_walker: ExpressionWalker):
         """
-        return list of unique views, expectation list is of size zero or one
+        Apply expression to argument.
         """
-        views = list()
-        return views
-
-    def replace_view(self, view):
-        """
-        Move expression to a new view.
-        """
-        return self
-
-    def evaluate(self, data_frame):
-        """
-        Evaluate expression, taking data from data_frame.
-        """
-        return self.value
+        assert isinstance(expr_walker, ExpressionWalker)
+        return expr_walker.act_on_literal(value=self.value)
 
     def to_python(self, *, want_inline_parens: bool = False) -> PythonText:
         """
@@ -1093,6 +1066,11 @@ class Value(Term):
         return Value(-self.value)
 
 
+def lit(x):
+    """Represent a value"""
+    return Value(x)
+
+
 class ListTerm(PreTerm):
     """
     Class to hold a collection.
@@ -1101,7 +1079,7 @@ class ListTerm(PreTerm):
     # derived from PreTerm as this is not combinable
     def __init__(self, value):
         assert isinstance(value, (list, tuple))
-        self.value = [v for v in value]  # copy and standardize to a list
+        self.value = list(value)  # copy and standardize to a list
         PreTerm.__init__(self)
 
     def is_equal(self, other):
@@ -1113,35 +1091,18 @@ class ListTerm(PreTerm):
             return False
         return self.value == other.value
 
-    def get_views(self):
+    def act_on(self, arg, *, expr_walker: ExpressionWalker):
         """
-        return list of unique views, expectation list is of size zero or one
-        """
-        views = list()
-        for ai in self.value:
-            if isinstance(ai, PreTerm):
-                vi = ai.get_views()
-                for vii in vi:
-                    if vii not in views:  # expect list to be of size zero or one
-                        views.append(vii)
-        return views
-
-    def replace_view(self, view):
-        """
-        Move expression to a new view.
-        """
-        new_list = [ai.replace_view(view) for ai in self.value]
-        return ListTerm(new_list)
-
-    def evaluate(self, data_frame):
-        """
-        Evaluate expression, taking data from data_frame.
+        Apply expression to argument.
         """
+        assert isinstance(expr_walker, ExpressionWalker)
         res = [None] * len(self.value)
         for i in range(len(self.value)):
             vi = self.value[i]
             if isinstance(vi, PreTerm):
-                vi = vi.evaluate(data_frame)
+                vi = vi.act_on(arg, expr_walker=expr_walker)
+            else:
+                vi = expr_walker.act_on_literal(value=vi)
             res[i] = vi
         return res
 
@@ -1194,24 +1155,12 @@ class DictTerm(PreTerm):
             return False
         return self.value == other.value
 
-    def get_views(self):
+    def act_on(self, arg, *, expr_walker: ExpressionWalker):
         """
-        return list of unique views, expectation list is of size zero or one
+        Apply expression to argument.
         """
-        views = list()
-        return views
-
-    def replace_view(self, view):
-        """
-        Move expression to a new view.
-        """
-        return DictTerm(self.value)
-
-    def evaluate(self, data_frame):
-        """
-        Evaluate expression, taking data from data_frame.
-        """
-        return self.value.copy()
+        assert isinstance(expr_walker, ExpressionWalker)
+        return expr_walker.act_on_literal(value=self.value.copy())
 
     def to_python(self, *, want_inline_parens: bool = False) -> PythonText:
         """
@@ -1249,29 +1198,19 @@ def enc_value(value):
 
 class ColumnReference(Term):
     """class to represent referring to a column"""
-
-    view: Optional[Any]
     column_name: str
 
-    def __init__(self, view, column_name):
-        self.view = view
+    def __init__(self, column_name):
         self.column_name = column_name
         assert isinstance(column_name, str)
-        # # an invariant we want, but don't want the expense of checking it against lists
-        # if view is not None:
-        #     if column_name not in view.column_names:
-        #         raise KeyError(
-        #             "column_name '"
-        #             + str(column_name)
-        #             + "' must be a column of the given view"
-        #         )
         Term.__init__(self)
 
-    def evaluate(self, data_frame):
+    def act_on(self, arg, *, expr_walker: ExpressionWalker):
         """
-        Evaluate expression, taking data from data_frame.
+        Apply expression to argument.
         """
-        return data_frame[self.column_name]
+        assert isinstance(expr_walker, ExpressionWalker)
+        return expr_walker.act_on_column_name(arg=arg, value=self.column_name)
 
     def is_equal(self, other):
         """
@@ -1280,24 +1219,8 @@ class ColumnReference(Term):
         # can't use == as that builds a larger expression
         if not isinstance(other, ColumnReference):
             return False
-        if self.view != other.view:
-            return False
         return self.column_name == other.column_name
 
-    def get_views(self):
-        """
-        return list of unique views, expectation list is of size zero or one
-        """
-        views = list()
-        views.append(self.view)
-        return views
-
-    def replace_view(self, view):
-        """
-        Move expression to a new view.
-        """
-        return ColumnReference(view=view, column_name=self.column_name)
-
     def to_python(self, *, want_inline_parens: bool = False) -> PythonText:
         """
         Convert parsed expression into a string
@@ -1316,6 +1239,12 @@ class ColumnReference(Term):
         columns_seen.add(self.column_name)
 
 
+def col(nm: str):
+    """represent a column or value"""
+    assert isinstance(nm, str)
+    return ColumnReference(column_name=nm)
+
+
 # noinspection SpellCheckingInspection
 def _can_find_method_by_name(op):
     assert isinstance(op, str)
@@ -1331,16 +1260,17 @@ def _can_find_method_by_name(op):
         return True
     # check user fns
     # first check chosen mappings
+    data_model = data_algebra.data_model.default_data_model()  # just use default for checking defs
     try:
         # noinspection PyUnusedLocal
-        check_val = data_algebra.pandas_model.default_data_model.user_fun_map[op]  # for KeyError
+        check_val = data_model.user_fun_map[op]  # for KeyError
         return True
     except KeyError:
         pass
     # check chosen mappings
     try:
         # noinspection PyUnusedLocal
-        check_val = data_algebra.pandas_model.default_data_model.impl_map[op]  # for KeyError
+        check_val = data_model.impl_map[op]  # for KeyError
         return True
     except KeyError:
         pass
@@ -1403,31 +1333,6 @@ class Expression(Term):
                 return False
         return True
 
-    def get_views(self):
-        """
-        return list of unique views, expectation list is of size zero or one
-        """
-        views = list()
-        for ai in self.args:
-            vi = ai.get_views()
-            for vii in vi:
-                if vii not in views:  # expect list to be of size zero or one
-                    views.append(vii)
-        return views
-
-    def replace_view(self, view):
-        """
-        Move expression to a new view.
-        """
-        new_args = [oi.replace_view(view) for oi in self.args]
-        return Expression(
-            op=self.op,
-            args=new_args,
-            params=self.params,
-            inline=self.inline,
-            method=self.method,
-        )
-
     def get_column_names(self, columns_seen: Set[str]) -> None:
         """
         Add column names to columns_seen
@@ -1446,46 +1351,14 @@ class Expression(Term):
         """
         methods_seen.add(self.op)
 
-    def evaluate(self, data_frame):
+    def act_on(self, arg, *, expr_walker: ExpressionWalker):
         """
-        Evaluate expression, taking data from data_frame.
+        Apply expression to argument.
         """
-        args = [ai.evaluate(data_frame) for ai in self.args]
-        # check user fns
-        # first check chosen mappings
-        try:
-            method_to_call = data_algebra.pandas_model.default_data_model.user_fun_map[self.op]
-            return method_to_call(*args)
-        except KeyError:
-            pass
-        # check chosen mappings
-        try:
-            method_to_call = data_algebra.pandas_model.default_data_model.impl_map[self.op]
-            return method_to_call(*args)
-        except KeyError:
-            pass
-        # special zero argument functions
-        if len(args) == 0:
-            if self.op == "_uniform":
-                return numpy.random.uniform(size=data_frame.shape[0])
-            else:
-                KeyError(f"zero-argument function {self.op} not found")
-        # now see if argument (usually Pandas) can do this
-        # doubt we hit in this, as most exposed methods are window methods
-        try:
-            method = getattr(args[0], self.op)
-            if callable(method):
-                return method(*args[1:])
-        except AttributeError:
-            pass
-        # new see if numpy can do this
-        try:
-            fn = numpy.__dict__[self.op]
-            if callable(fn):
-                return fn(*args)
-        except KeyError:
-            pass
-        raise KeyError(f"function {self.op} not found")
+        assert isinstance(expr_walker, ExpressionWalker)
+        args = [ai.act_on(arg, expr_walker=expr_walker) for ai in self.args]
+        res = expr_walker.act_on_expression(arg=arg, values=args, op=self)
+        return res
 
     def to_python(self, *, want_inline_parens: bool = False) -> PythonText:
         """
