@@ -20,11 +20,15 @@ import data_algebra.expression_walker
 
 
 def _build_lit(v):
-    if isinstance(v, int):
+    if isinstance(v, bool):
+        # bools return true to isinstance int.
+        return pl.lit(v)
+    elif isinstance(v, int):
         # Polars defaults ints in constructor to Int64,
         # but ints in lit to Int32. Try to prevent type clashes
         return pl.lit(v, pl.Int64)
-    return pl.lit(v)
+    else:
+        return pl.lit(v)
 
 
 def _reduce_plus(*args):
@@ -63,13 +67,58 @@ _da_temp_zero_column_name = "_da_temp_zero_column"
 _da_temp_one_column_name = "_da_temp_one_column"
 
 
-class ExpressionRequirements(data_algebra.expression_walker.ExpressionWalker):
+class PolarsTerm:
+    """
+    Class to carry Polars expression term and annotations about expression tree.
+    """
+
+    polars_term: Any
+    lit_value: Any
+    is_literal: bool
+    is_column: bool
+    is_series: bool
+
+    def __init__(
+        self, 
+        *, 
+        polars_term = None, 
+        is_literal: bool = False,
+        is_column: bool = False,
+        is_series: bool = False,
+        lit_value = None,
+        ) -> None:
+        """
+        Carry a Polars expression term (polars_term) plus annotations.
+
+        :param polars_term: Optional Polars expression (None means collect info, not a true term)
+        :param is_literal: True if term is a constant
+        :param is_column: True if term is a column name
+        :param lit_value: original value for a literal
+        :param inputs: inputs to expression node
+        """
+        assert isinstance(is_literal, bool)
+        assert isinstance(is_column, bool)
+        assert isinstance(is_series, bool)
+        assert (is_literal + is_column + is_series) <= 1
+        if lit_value is not None:
+            assert is_literal
+        self.lit_value = lit_value
+        self.polars_term = polars_term
+        self.is_literal = is_literal
+        self.is_column = is_column
+        self.is_series = is_series
+
+
+class ExpressionRequirementsCollector(data_algebra.expression_walker.ExpressionWalker):
     """
     Class to collect what accommodations an expression needs.
     """
     collect_required: bool
     zero_constant_required: bool
     one_constant_required: bool
+    _needs_zero_constant: Set[str]
+    _needs_one_constant: Set[str]
+    _collect_required: Set[str]
 
     def __init__(self) -> None:
         data_algebra.expression_walker.ExpressionWalker.__init__(
@@ -88,7 +137,8 @@ class ExpressionRequirements(data_algebra.expression_walker.ExpressionWalker):
         }
         self._collect_required = {
             "uniform", "_uniform",
-            "ngroup", "_ngroup",
+            "ngroup", "_ngroup",  # number groups
+            "sgroup", "_sgroup",  # size groups
         }
     
     def act_on_literal(self, *, value):
@@ -141,42 +191,6 @@ class ExpressionRequirements(data_algebra.expression_walker.ExpressionWalker):
             temp_v_columns.append(_build_lit(1).alias(_da_temp_one_column_name))
 
 
-class PolarsTerm:
-    """
-    Class to carry Polars expression term and annotations about expression tree.
-    """
-
-    polars_term: Any
-    lit_value: Any
-    is_literal: bool
-    is_column: bool
-
-    def __init__(
-        self, 
-        *, 
-        polars_term = None, 
-        is_literal: bool = False,
-        is_column: bool = False,
-        lit_value = None,
-        ) -> None:
-        """
-        Carry a Polars expression term (polars_term) plus annotations.
-
-        :param polars_term: Optional Polars expression (None means collect info, not a true term)
-        :param is_literal: True if term is a constant
-        :param is_column: True if term is a column name
-        :param lit_value: original value for a literal
-        :param inputs: inputs to expression node
-        """
-        assert isinstance(is_literal, bool)
-        assert isinstance(is_column, bool)
-        if lit_value is not None:
-            assert is_literal
-        self.lit_value = lit_value
-        self.polars_term = polars_term
-        self.is_literal = is_literal
-
-
 def _unpack_lits(v):
     if isinstance(v, PolarsTerm):
         if v.is_literal:
@@ -198,22 +212,36 @@ def _mapv(a, b: Dict, c):
     return res
 
 
-def _populate_expr_impl_map() -> Dict[int, Dict[str, Callable]]:
+def _populate_expr_impl_map(extend_context: bool) -> Dict[int, Dict[str, Callable]]:
     """
     Map symbols to implementations.
     """
+    assert isinstance(extend_context, bool)
     # TODO: fill in more
-    impl_map_0 = {
-        "count": lambda : pl.col(_da_temp_one_column_name).cumsum(),  # ugly SQL def
-        "_count": lambda : pl.col(_da_temp_one_column_name).cumsum(),  # ugly SQL def
-        "cumcount": lambda : pl.col(_da_temp_one_column_name).cumsum(),
-        "_cumcount": lambda : pl.col(_da_temp_one_column_name).cumsum(),
-        "row_number": lambda : pl.col(_da_temp_one_column_name).cumsum(),
-        "_row_number": lambda : pl.col(_da_temp_one_column_name).cumsum(),
-        "size": lambda : pl.col(_da_temp_one_column_name).sum(),
-        "_size": lambda : pl.col(_da_temp_one_column_name).sum(),
-    }
+    if extend_context:
+        impl_map_0 = {
+            "count": lambda : pl.col(_da_temp_one_column_name).cumsum(),  # ugly SQL def
+            "_count": lambda : pl.col(_da_temp_one_column_name).cumsum(),  # ugly SQL def
+            "cumcount": lambda : pl.col(_da_temp_one_column_name).cumsum(),
+            "_cumcount": lambda : pl.col(_da_temp_one_column_name).cumsum(),
+            "row_number": lambda : pl.col(_da_temp_one_column_name).cumsum(),
+            "_row_number": lambda : pl.col(_da_temp_one_column_name).cumsum(),
+            "size": lambda : pl.col(_da_temp_one_column_name).sum(),
+            "_size": lambda : pl.col(_da_temp_one_column_name).sum(),
+        }
+    else:
+        impl_map_0 = {
+            "count": lambda : pl.col(_da_temp_one_column_name).sum(),
+            "_count": lambda : pl.col(_da_temp_one_column_name).sum(),
+            "cumcount": lambda : pl.col(_da_temp_one_column_name).sum(),
+            "_cumcount": lambda : pl.col(_da_temp_one_column_name).sum(),
+            "row_number": lambda : pl.col(_da_temp_one_column_name).sum(),
+            "_row_number": lambda : pl.col(_da_temp_one_column_name).sum(),
+            "size": lambda : pl.col(_da_temp_one_column_name).sum(),
+            "_size": lambda : pl.col(_da_temp_one_column_name).sum(),
+        }
     impl_map_1 = {
+        "+": lambda x: x,
         "-": lambda x: 0 - x,
         "abs": lambda x: x.abs(),
         "all": lambda x: x.all(),
@@ -234,8 +262,8 @@ def _populate_expr_impl_map() -> Dict[int, Dict[str, Callable]]:
         "coalesce0": lambda x: pl.when(x.is_null()).then(pl.col(_da_temp_zero_column_name)).otherwise(x),
         "cos": lambda x: x.cos(),
         "cosh": lambda x: x.cosh(),
-        "count": lambda x: pl.col(_da_temp_one_column_name).cumsum(),
-        "cumcount": lambda x: pl.col(_da_temp_one_column_name).cumsum(),
+        "count": lambda x: pl.when(x.is_null() | x.is_nan()).then(_build_lit(0)).otherwise(_build_lit(1)).sum(),  # not tested yet TODO
+        "cumcount": lambda x: pl.when(x.is_null() | x.is_nan()).then(_build_lit(0)).otherwise(_build_lit(1)).cumsum(),  # not working yet TODO
         "cummax": lambda x: x.cummax(),
         "cummin": lambda x: x.cummin(),
         "cumprod": lambda x: x.cumprod(),
@@ -324,7 +352,148 @@ def _populate_expr_impl_map() -> Dict[int, Dict[str, Callable]]:
     return impl_map
 
 
-class PolarsModel(data_algebra.data_model.DataModel, data_algebra.expression_walker.ExpressionWalker):
+class PolarsExpressionActor(data_algebra.expression_walker.ExpressionWalker):
+    """Act on expressions in Polars context"""
+
+    extend_context: bool
+    project_context: bool
+    partition_by: List[str]
+
+    def __init__(
+        self, 
+        *, 
+        polars_model, 
+        extend_context: bool = False, 
+        project_context: bool = False,
+        partition_by: Optional[Iterable[str]] = None
+        ) -> None:
+        assert isinstance(extend_context, bool)
+        assert isinstance(project_context, bool)
+        assert (extend_context + project_context) == 1
+        data_algebra.expression_walker.ExpressionWalker.__init__(
+            self,
+        )
+        self.polars_model = polars_model
+        self.extend_context = extend_context
+        self.project_context = project_context
+        assert not isinstance(partition_by, str)  # common error
+        if partition_by is None:
+            partition_by = []
+        else:
+            partition_by = list(partition_by)
+        self.partition_by = partition_by
+
+   # expression helpers
+    
+    def act_on_literal(self, *, value):
+        """
+        Action for a literal/constant in an expression.
+
+        :param value: literal value being supplied
+        :return: converted result
+        """
+        assert not isinstance(value, PolarsTerm)
+        if isinstance(value, (Dict, List, Set)):
+            return PolarsTerm(polars_term=None, is_literal=True, lit_value=value)
+        else:
+            return PolarsTerm(polars_term=_build_lit(value), is_literal=True, lit_value=value)
+    
+    def act_on_column_name(self, *, arg, value):
+        """
+        Action for a column name.
+
+        :param arg: None
+        :param value: column name
+        :return: arg acted on
+        """
+        assert isinstance(arg, (pl.DataFrame, type(None)))
+        assert isinstance(value, str)
+        return PolarsTerm(polars_term=pl.col(value), is_column=True)
+    
+    def act_on_expression(self, *, arg, values: List, op):
+        """
+        Action for a column name.
+
+        :param arg: None
+        :param values: list of values to work on
+        :param op: operator to apply
+        :return: arg acted on
+        """
+        assert isinstance(arg, (pl.DataFrame, type(None)))
+        assert isinstance(values, List)
+        assert isinstance(op, data_algebra.expr_rep.Expression)
+        # process inputs
+        for v in values:
+            assert isinstance(v, (List, PolarsTerm))
+        want_literals_unpacked = (op.op in self.polars_model.want_literals_unpacked)
+        if want_literals_unpacked:
+            args = _unpack_lits(values)
+        else:
+            args = [v.polars_term for v in values]
+        # lookup method
+        f = None
+        arity = len(values)
+        if (f is None) and (arity == 0):
+            if op.op in ["_uniform", "uniform"]:
+                assert isinstance(arg, pl.DataFrame)
+                return PolarsTerm(
+                    polars_term=pl.Series(
+                        values=self.polars_model.rng.uniform(0.0, 1.0, arg.shape[0]),
+                        dtype=pl.datatypes.Float64,
+                        dtype_if_empty=pl.datatypes.Float64),
+                    is_series=True,
+                )
+            elif op.op in ["_sgroup", "sgroup"]:
+                assert isinstance(arg, pl.DataFrame)
+                n_groups = 0
+                if arg.shape[0] > 0:
+                    n_groups = 1
+                    if len(self.partition_by) > 0:
+                        s_groups = arg.groupby(self.partition_by).apply(lambda x: x.head(1)).shape[0]
+                return PolarsTerm(
+                    polars_term=pl.lit(s_groups),
+                    lit_value=n_groups,
+                    is_literal=True,
+                )
+            elif op.op in ["_ngroup", "ngroup"]:
+                assert isinstance(arg, pl.DataFrame)
+                group_labels = []
+                if arg.shape[0] > 0:
+                    n_groups = [0] * arg.shape[0]
+                    if len(self.partition_by) > 0:
+                        # TODO: number the groups, not size them
+                        n_groups = arg.groupby(self.partition_by).apply(lambda x: x.head(1)).shape[0]
+                return PolarsTerm(
+                    polars_term=pl.Series(
+                        values=group_labels,
+                        dtype=pl.datatypes.Int64,
+                        dtype_if_empty=pl.datatypes.Int64),
+                    is_series=True,
+                )
+        if f is None:
+            try:
+                if self.extend_context:
+                    f = self.polars_model.extend_expr_impl_map[len(values)][op.op]
+                elif self.project_context:
+                    f = self.polars_model.project_expr_impl_map[len(values)][op.op]
+            except KeyError:
+                pass
+        if (f is None) and (arity > 0):
+            try:
+                f = self.polars_model.impl_map_arbitrary_arity[op.op]
+            except KeyError:
+                pass
+        if f is None:
+            raise ValueError(f"failed to lookup {op}")
+        # apply method
+        res = f(*args)
+        # wrap result
+        return PolarsTerm(
+            polars_term=res,
+        )
+
+
+class PolarsModel(data_algebra.data_model.DataModel):
     """
     Interface for realizing the data algebra as a sequence of steps over Polars https://www.pola.rs .
 
@@ -334,19 +503,18 @@ class PolarsModel(data_algebra.data_model.DataModel, data_algebra.expression_wal
     use_lazy_eval: bool
     presentation_model_name: str
     _method_dispatch_table: Dict[str, Callable]
-    _expr_impl_map: Dict[int, Dict[str, Callable]]
-    _impl_map_arbitrary_arity: Dict[str, Callable]
+    extend_expr_impl_map: Dict[int, Dict[str, Callable]]
+    project_expr_impl_map: Dict[int, Dict[str, Callable]]
+    impl_map_arbitrary_arity: Dict[str, Callable]
     _collect_required: Set[str]
-    _rng: Any
+    want_literals_unpacked: Set[str]
+    rng: Any
 
     def __init__(self, *, use_lazy_eval: bool = True):
         data_algebra.data_model.DataModel.__init__(
             self, presentation_model_name="pl", module=pl
         )
-        data_algebra.expression_walker.ExpressionWalker.__init__(
-            self,
-        )
-        self._rng = np.random.default_rng()
+        self.rng = np.random.default_rng()
         assert isinstance(use_lazy_eval, bool)
         self.use_lazy_eval = use_lazy_eval
         self._method_dispatch_table = {
@@ -364,8 +532,9 @@ class PolarsModel(data_algebra.data_model.DataModel, data_algebra.expression_wal
             "SQLNode": self._sql_proxy_step,
             "TableDescription": self._table_step,
         }
-        self._expr_impl_map = _populate_expr_impl_map()
-        self._impl_map_arbitrary_arity = {
+        self.extend_expr_impl_map = _populate_expr_impl_map(extend_context=True)
+        self.project_expr_impl_map = _populate_expr_impl_map(extend_context=False)
+        self.impl_map_arbitrary_arity = {
             "concat": lambda *args: pl.concat_str(args),
             "fmax": lambda *args: pl.max(args),
             "fmin": lambda *args: pl.min(args),
@@ -378,7 +547,7 @@ class PolarsModel(data_algebra.data_model.DataModel, data_algebra.expression_wal
             "or": _reduce_or,
             "|": _reduce_or,
         }
-        self._want_literals_unpacked = {
+        self.want_literals_unpacked = {
             "around",
             "is_in",
             "mapv",
@@ -461,6 +630,15 @@ class PolarsModel(data_algebra.data_model.DataModel, data_algebra.expression_wal
             return frame_list[0]
         res = pl.concat(frame_list, how="horizontal")
         return res
+    
+    def get_cell(self, *, d, row: int, colname: str):
+        """get a value from a cell"""
+        return d[row, colname]
+    
+    def set_col(self, *, d, colname: str, values):
+        """set column, return ref"""
+        d2 = d.with_column(pl.Series(values=values).alias(colname))
+        return d2
 
     def table_is_keyed_by_columns(self, table, *, column_names: Iterable[str]) -> bool:
         """
@@ -470,6 +648,11 @@ class PolarsModel(data_algebra.data_model.DataModel, data_algebra.expression_wal
         :param column_names: list of column names
         :return: True if rows are uniquely keyed by values in named columns
         """
+        # get rid of some corner cases
+        if table.shape[0] < 2:
+            return True
+        if (column_names is None):
+            return False
         # check for ill-condition
         if isinstance(column_names, str):
             column_names = [column_names]
@@ -479,8 +662,6 @@ class PolarsModel(data_algebra.data_model.DataModel, data_algebra.expression_wal
         if len(missing_columns) > 0:
             return False
         # get rid of some corner cases
-        if table.shape[0] < 2:
-            return True
         if len(column_names) < 1:
             return False
         mx = (
@@ -570,12 +751,12 @@ class PolarsModel(data_algebra.data_model.DataModel, data_algebra.expression_wal
         partition_by = op.partition_by
         temp_v_columns = []
         # see if we need to make partition non-empty
-        if len(partition_by) <= 0:
+        if (partition_by is None) or (len(partition_by) <= 0):
             v_name = "_da_extend_temp_partition_column"
             partition_by = [v_name]
             temp_v_columns.append(_build_lit(1).alias(v_name))
         # pre-scan expressions
-        er = ExpressionRequirements()
+        er = ExpressionRequirementsCollector()
         for opk in op.ops.values():
             opk.act_on(None, expr_walker=er)
         er.add_in_temp_columns(temp_v_columns)
@@ -601,10 +782,12 @@ class PolarsModel(data_algebra.data_model.DataModel, data_algebra.expression_wal
                         inline=opk.inline,
                         method=opk.method,
                     )
-            fld_k_container = opk.act_on(value_to_send_to_act, expr_walker=self)  # PolarsTerm
+            fld_k_container = opk.act_on(
+                value_to_send_to_act, 
+                expr_walker=PolarsExpressionActor(polars_model=self, extend_context=True, partition_by=op.partition_by))  # PolarsTerm
             assert isinstance(fld_k_container, PolarsTerm)
             fld_k = fld_k_container.polars_term
-            if op.windowed_situation:
+            if op.windowed_situation and (not (fld_k_container.is_literal or fld_k_container.is_column or fld_k_container.is_series)):
                 fld_k = fld_k.over(partition_by)
             produced_columns.append(fld_k.alias(k))
         if len(temp_v_columns) > 0:
@@ -620,6 +803,7 @@ class PolarsModel(data_algebra.data_model.DataModel, data_algebra.expression_wal
         res = res.with_columns(produced_columns)
         if len(temp_v_columns) > 0:
             res = res.select(op.columns_produced())
+        # get back to lazy type if needed
         if self.use_lazy_eval and isinstance(res, pl.DataFrame):
             res = res.lazy()
         return res
@@ -639,7 +823,7 @@ class PolarsModel(data_algebra.data_model.DataModel, data_algebra.expression_wal
             group_by = [v_name]
             temp_v_columns.append(_build_lit(1).alias(v_name))
         # pre-scan expressions
-        er = ExpressionRequirements()
+        er = ExpressionRequirementsCollector()
         for opk in op.ops.values():
             opk.act_on(None, expr_walker=er)
         er.add_in_temp_columns(temp_v_columns)
@@ -664,7 +848,9 @@ class PolarsModel(data_algebra.data_model.DataModel, data_algebra.expression_wal
                     inline=opk.inline, 
                     method=opk.method,
                 )
-            fld_k_container = opk.act_on(value_to_send_to_act, expr_walker=self)  # PolarsTerm
+            fld_k_container = opk.act_on(
+                value_to_send_to_act, 
+                expr_walker=PolarsExpressionActor(polars_model=self, project_context=True))  # PolarsTerm
             assert isinstance(fld_k_container, PolarsTerm)
             fld_k = fld_k_container.polars_term
             produced_columns.append(fld_k.alias(k))
@@ -673,6 +859,14 @@ class PolarsModel(data_algebra.data_model.DataModel, data_algebra.expression_wal
         res = res.groupby(group_by).agg(produced_columns)
         if len(temp_v_columns) > 0:
             res = res.select(op.columns_produced())
+        if (op.group_by is None) or (len(op.group_by) == 0):
+            # see if we have a zero row result
+            if isinstance(res, pl.LazyFrame):
+                res = res.collect()
+            if res.shape[0] <= 0:
+                # make an all None frame
+                res = pl.DataFrame({c: [None] for c in res.columns}, columns=[(res.columns[j], res.dtypes[j]) for j in range(res.shape[1])])
+        # see if we need to convert to lazy type
         if self.use_lazy_eval and isinstance(res, pl.DataFrame):
             res = res.lazy()
         return res
@@ -690,12 +884,18 @@ class PolarsModel(data_algebra.data_model.DataModel, data_algebra.expression_wal
         how = op.jointype.lower()
         if how == "full":
             how = "outer"
-        coalesce_columns = (
-            set(op.sources[0].columns_produced()).intersection(op.sources[1].columns_produced()) 
-            - set(op.on_a))
         if how != "right":
+            coalesce_columns = (
+                set(op.sources[0].columns_produced()).intersection(op.sources[1].columns_produced()) 
+                - set(op.on_a))
+            orphan_keys = [c for c in op.on_b if c not in set(op.on_a)]
+            input_right = inputs[1]
+            if len(orphan_keys) > 0:
+                input_right = input_right.with_columns([
+                    pl.col(c).alias(f"{c}_da_join_tmp_key") for c in orphan_keys
+                ])
             res = inputs[0].join(
-                inputs[1],
+                input_right,
                 left_on=op.on_a,
                 right_on=op.on_b,
                 how=how,
@@ -709,10 +909,21 @@ class PolarsModel(data_algebra.data_model.DataModel, data_algebra.expression_wal
                         .alias(c)
                     for c in coalesce_columns
                 ])
+            if len(orphan_keys) > 0:
+                res = res.rename({f"{c}_da_join_tmp_key": c for c in orphan_keys})
         else:
             # simulate right join with left join
+            coalesce_columns = (
+                set(op.sources[0].columns_produced()).intersection(op.sources[1].columns_produced()) 
+                - set(op.on_b))
+            orphan_keys = [c for c in op.on_a if c not in set(op.on_b)]
+            input_right = inputs[0]
+            if len(orphan_keys) > 0:
+                input_right = input_right.with_columns([
+                    pl.col(c).alias(f"{c}_da_join_tmp_key") for c in orphan_keys
+                ])
             res = inputs[1].join(
-                inputs[0],
+                input_right,
                 left_on=op.on_b,
                 right_on=op.on_a,
                 how="left",
@@ -726,6 +937,8 @@ class PolarsModel(data_algebra.data_model.DataModel, data_algebra.expression_wal
                         .alias(c)
                     for c in coalesce_columns
                 ])
+            if len(orphan_keys) > 0:
+                res = res.rename({f"{c}_da_join_tmp_key": c for c in orphan_keys})
         res = res.select(op.columns_produced())
         return res
     
@@ -811,7 +1024,7 @@ class PolarsModel(data_algebra.data_model.DataModel, data_algebra.expression_wal
         res = self._compose_polars_ops(op.sources[0], data_map=data_map)
         temp_v_columns = []
         # pre-scan expressions
-        er = ExpressionRequirements()
+        er = ExpressionRequirementsCollector()
         for opk in op.ops.values():
             opk.act_on(None, expr_walker=er)
         er.add_in_temp_columns(temp_v_columns)
@@ -823,7 +1036,9 @@ class PolarsModel(data_algebra.data_model.DataModel, data_algebra.expression_wal
         # work on expression
         if len(temp_v_columns) > 0:
             res = res.with_columns(temp_v_columns)
-        selection = op.expr.act_on(value_to_send_to_act, expr_walker=self)  # PolarsTerm
+        selection = op.expr.act_on(
+            value_to_send_to_act, 
+            expr_walker=PolarsExpressionActor(polars_model=self, extend_context=True))  # PolarsTerm
         assert isinstance(selection, PolarsTerm)
         res = res.filter(selection.polars_term)
         if len(temp_v_columns) > 0:
@@ -881,7 +1096,7 @@ class PolarsModel(data_algebra.data_model.DataModel, data_algebra.expression_wal
         assert self.is_appropriate_data_instance(data)
         assert isinstance(data, pl.DataFrame)
         assert blocks_in is not None
-        assert blocks_in.control_table.shape[0] > 1
+        assert blocks_in.control_table.shape[0] > 0
         assert len(blocks_in.control_table_keys) > 0
         data = data.select(blocks_in.block_columns)
         assert set(data.columns) == set(blocks_in.block_columns)
@@ -950,7 +1165,7 @@ class PolarsModel(data_algebra.data_model.DataModel, data_algebra.expression_wal
         assert self.is_appropriate_data_instance(data)
         assert isinstance(data, pl.DataFrame)
         assert blocks_out is not None
-        assert blocks_out.control_table.shape[0] > 1
+        assert blocks_out.control_table.shape[0] > 0
         assert len(blocks_out.control_table_keys) > 0
         data = data.select(blocks_out.row_columns)
         assert set(data.columns) == set(blocks_out.row_columns)
@@ -986,89 +1201,6 @@ class PolarsModel(data_algebra.data_model.DataModel, data_algebra.expression_wal
         else:
             res = res.sort(blocks_out.control_table_keys)
         return res
-    
-    # expression helpers
-    
-    def act_on_literal(self, *, value):
-        """
-        Action for a literal/constant in an expression.
-
-        :param value: literal value being supplied
-        :return: converted result
-        """
-        assert not isinstance(value, PolarsTerm)
-        if isinstance(value, (Dict, List, Set)):
-            return PolarsTerm(polars_term=None, is_literal=True, lit_value=value)
-        else:
-            return PolarsTerm(polars_term=_build_lit(value), is_literal=True, lit_value=value)
-    
-    def act_on_column_name(self, *, arg, value):
-        """
-        Action for a column name.
-
-        :param arg: None
-        :param value: column name
-        :return: arg acted on
-        """
-        assert isinstance(arg, (pl.DataFrame, type(None)))
-        assert isinstance(value, str)
-        return PolarsTerm(polars_term=pl.col(value), is_column=True)
-    
-    def act_on_expression(self, *, arg, values: List, op):
-        """
-        Action for a column name.
-
-        :param arg: None
-        :param values: list of values to work on
-        :param op: operator to apply
-        :return: arg acted on
-        """
-        assert isinstance(arg, (pl.DataFrame, type(None)))
-        assert isinstance(values, List)
-        assert isinstance(op, data_algebra.expr_rep.Expression)
-        # process inputs
-        for v in values:
-            assert isinstance(v, (List, PolarsTerm))
-        want_literals_unpacked = (op.op in self._want_literals_unpacked)
-        if want_literals_unpacked:
-            args = _unpack_lits(values)
-        else:
-            args = [v.polars_term for v in values]
-        # lookup method
-        f = None
-        arity = len(values)
-        if (f is None) and (arity == 0):
-            if op.op in ["_uniform", "uniform"]:
-                assert isinstance(arg, pl.DataFrame)
-                return PolarsTerm(
-                    polars_term=pl.Series(
-                        values=self._rng.uniform(0.0, 1.0, arg.shape[0]),
-                        dtype=pl.datatypes.Float64,
-                        dtype_if_empty=pl.datatypes.Float64),
-                )
-        if (f is None): 
-            if op.op in ["_ngroup", "ngroup"]:
-                assert isinstance(arg, pl.DataFrame)
-                # n_groups = arg.groupby(["x"]).apply(lambda x: x.head(1)).shape[0]
-                raise ValueError(f" {op.op} not implemented for Polars adapter, yet")
-        if f is None:
-            try:
-                f = self._expr_impl_map[len(values)][op.op]
-            except KeyError:
-                pass
-        if (f is None) and (arity > 0):
-            try:
-                f = self._impl_map_arbitrary_arity[op.op]
-            except KeyError:
-                pass
-        if f is None:
-            raise ValueError(f"failed to lookup {op}")
-        # apply method
-        res = f(*args)
-        # wrap result
-        return PolarsTerm(
-            polars_term=res,
-        )
 
 
 def register_polars_model(key:Optional[str] = None):
