@@ -5,7 +5,7 @@ data algebra solutions to common data processing problems
 from typing import Any, Dict, Iterable, Optional, Tuple
 
 import numpy
-import data_algebra.pandas_model
+import data_algebra.data_model
 from data_algebra.data_ops import descr, TableDescription, ViewRepresentation
 from data_algebra.cdata import (
     pivot_specification,
@@ -224,28 +224,33 @@ def xicor_score_variables_plan(
     assert isinstance(n_rep, int)
     record_map = RecordMap(
         blocks_out=RecordSpecification(
-            control_table=data_algebra.pandas_model.default_data_model.pd.DataFrame(
+            control_table=data_algebra.data_model.default_data_model().data_frame(
                 {
                     "variable_name": x_vars,
                     "x": x_vars,
                     "y": y_name,
                 }
             ),
-            record_keys=[],
+            record_keys=["_da_xicor_tmp_index"],
             control_table_keys=["variable_name"],
-        )
+            strict=False,
+        ),
+        strict=False,
     )
-    rep_frame = data_algebra.pandas_model.default_data_model.pd.DataFrame({"rep": range(n_rep)})
+    rep_frame = data_algebra.data_model.default_data_model().data_frame({"rep": range(n_rep)})
     grouped_calc = (
         xicor_query(
-            d.convert_records(
-                record_map
-            ).natural_join(  # cross join rows to get experiment repetitions
-                b=descr(rep_frame=rep_frame),
-                on=[],
-                jointype="cross",
-            ),
-            var_keys=["variable_name", "rep"],
+            d
+                .extend({"_da_xicor_tmp_order": "1"})
+                .extend({"_da_xicor_tmp_index": "(1).cumsum()"}, order_by=["_da_xicor_tmp_order"])
+                .convert_records(
+                    record_map
+                ).natural_join(  # cross join rows to get experiment repetitions
+                    b=descr(rep_frame=rep_frame),
+                    on=[],
+                    jointype="cross",
+                ),
+                var_keys=["variable_name", "rep"],
         )
         .project(
             {
@@ -336,111 +341,6 @@ def last_observed_carried_forward(
     return ops
 
 
-def braid_data(
-    *,
-    d_state: ViewRepresentation,
-    d_event: ViewRepresentation,
-    order_by: Iterable[str],
-    partition_by: Optional[Iterable[str]] = None,
-    state_value_column_name: str,
-    event_value_column_names: Iterable[str],
-    source_id_column: str = "record_type",
-    state_row_mark: str = "state_row",
-    event_row_mark: str = "event_row",
-    stand_in_values: Dict,
-    locf_to_use_column_name: str = "locf_to_use",
-    locf_non_null_rank_column_name: str = "locf_non_null_rank",
-    locf_tiebreaker_column_name: str = "locf_tiebreaker",
-) -> ViewRepresentation:
-    """
-    Mix data from two sources, ordering by order_by columns and carrying forward observations
-    on d_state value column.
-
-    :param d_state: ViewRepresentation representation of state by order_by.
-    :param d_event: ViewRepresentation representation of events by order_by.
-    :param order_by: columns to order by (non empty list of column names)
-    :param partition_by: optional partitioning column names
-    :param state_value_column_name: column to copy from d_state and propagate forward
-    :param event_value_column_names: columns to copy from d_event
-    :param source_id_column: name for source identification column.
-    :param state_row_mark: source annotation of state rows.
-    :param event_row_mark: source annotation of event rows.
-    :param stand_in_values: dictionary stand in values to use for state_value_column_name and event_value_column_names
-            needed to get column types correct, replaced by None and not passed further.
-    :param locf_to_use_column_name: name for a temporary values column
-    :param locf_non_null_rank_column_name: name for a temporary values column
-    :param locf_tiebreaker_column_name: name for a temporary values column
-    :return: ops
-    """
-    assert isinstance(d_state, ViewRepresentation)
-    assert isinstance(d_event, ViewRepresentation)
-    assert not isinstance(order_by, str)
-    order_by = list(order_by)
-    assert len(order_by) > 0
-    if partition_by is not None:
-        assert not isinstance(partition_by, str)
-        partition_by = list(partition_by)
-    else:
-        partition_by = []
-    assert isinstance(state_value_column_name, str)
-    assert not isinstance(event_value_column_names, str)
-    event_value_column_names = list(event_value_column_names)
-    assert isinstance(source_id_column, str)
-    assert isinstance(state_row_mark, str)
-    assert isinstance(event_row_mark, str)
-    assert isinstance(locf_to_use_column_name, str)
-    assert isinstance(locf_non_null_rank_column_name, str)
-    assert isinstance(locf_tiebreaker_column_name, str)
-    assert isinstance(stand_in_values, dict)
-    together = (
-        d_state.extend({k: stand_in_values[k] for k in event_value_column_names})
-        .select_columns(
-            partition_by
-            + order_by
-            + [state_value_column_name]
-            + event_value_column_names
-        )
-        .concat_rows(
-            b=(
-                d_event.extend(
-                    {state_value_column_name: stand_in_values[state_value_column_name]}
-                ).select_columns(
-                    partition_by
-                    + order_by
-                    + [state_value_column_name]
-                    + event_value_column_names
-                )
-            ),
-            id_column=source_id_column,
-            a_name=state_row_mark,
-            b_name=event_row_mark,
-        )
-        # clear out stand-in values
-        .extend(
-            {
-                state_value_column_name: f'({source_id_column} == "{event_row_mark}").if_else(None, {state_value_column_name})'
-            }
-        )
-        .extend(
-            {
-                k: f'({source_id_column} == "{state_row_mark}").if_else(None, {k})'
-                for k in event_value_column_names
-            }
-        )
-    )
-    ops = last_observed_carried_forward(
-        together,
-        order_by=order_by,
-        partition_by=partition_by,
-        value_column_name=state_value_column_name,
-        selection_predicate="is_null()",
-        locf_to_use_column_name=locf_to_use_column_name,
-        locf_non_null_rank_column_name=locf_non_null_rank_column_name,
-        locf_tiebreaker_column_name=locf_tiebreaker_column_name,
-    )
-    return ops
-
-
 def rank_to_average(
     d: ViewRepresentation,
     *,
@@ -523,13 +423,13 @@ def replicate_rows_query(
     assert power_key_colname != count_column_name
     assert power_key_colname not in d.column_names
     # get a pandas namespace
-    pd = data_algebra.pandas_model.default_data_model.pd
+    local_data_model = data_algebra.data_model.default_data_model()
     # build powers of 2 until max_count is met or exceeded
     powers = list(range(int(numpy.ceil(numpy.log(max_count) / numpy.log(2))) + 1))
     # replicate each power the number of times it specifies
-    count_frame = pd.concat(
+    count_frame = local_data_model.concat_rows(
         [
-            pd.DataFrame(
+            local_data_model.data_frame(
                 {
                     power_key_colname: f"p{p}",
                     seq_column_name: range(int(2**p)),
@@ -538,7 +438,7 @@ def replicate_rows_query(
             for p in powers
         ]
     )
-    count_frame.reset_index(drop=True, inplace=True)
+    local_data_model.clean_copy(count_frame)
     # specify ops that produce row replicates
     ops = (
         d
