@@ -4,11 +4,11 @@ Utils that help with testing. This module is allowed to import many other module
 
 import pickle
 import traceback
-from typing import Any, Optional
+from typing import Any, Dict, Optional
 
 import numpy
 
-import data_algebra.pandas_model
+import data_algebra.data_model
 import data_algebra.eval_cache
 import data_algebra.db_model
 import data_algebra.SQLite
@@ -21,6 +21,14 @@ from data_algebra.sql_format_options import SQLFormatOptions
 from data_algebra.data_ops import *
 
 
+have_polars = False
+try:
+    import data_algebra.polars_model  # conditional import
+    have_polars = True
+except ModuleNotFoundError:
+    pass
+
+
 # controls
 test_PostgreSQL = False  # causes an external dependency
 test_BigQuery = False  # causes an external dependency
@@ -33,7 +41,7 @@ run_direct_ops_path_tests = False
 global_test_result_cache: Optional[data_algebra.eval_cache.ResultCache] = None
 
 
-def re_parse(ops):
+def _re_parse(ops, *, data_model_map: Dict[str, Any]):
     """
     Return copy of object made by dumping to string via repr() and then evaluating that string.
     """
@@ -41,22 +49,27 @@ def re_parse(ops):
     ops2 = eval(
         str1,
         globals(),
-        {
-            "pd": data_algebra.pandas_model.default_data_model.pd
-        },  # make our definition of pandas available
+        data_model_map,  # make our definition of data module available
+            # cdata uses this
     )
     return ops2
 
 
-def formats_to_self(ops) -> bool:
+def formats_to_self(ops, *, data_model_map: Optional[Dict[str, Any]] = None) -> bool:
     """
     Check a operator dag formats and parses back to itself.
     Can raise exceptions. Also checks pickling.
 
     :param ops: data_algebra.data_ops.ViewRepresentation
+    :param data_model_map: map from abbreviated module names to modules (i.e. define pd)
     :return: logical, True if formats and evals back to self
     """
-    ops2 = re_parse(ops)
+    if data_model_map is None:
+        local_data_model = data_algebra.data_model.default_data_model()
+        data_model_map = {local_data_model.presentation_model_name: local_data_model.module}
+    ops2 = _re_parse(
+        ops, 
+        data_model_map=data_model_map)
     ops_match = ops == ops2
     assert ops_match
     pickle_string = pickle.dumps(ops)
@@ -73,13 +86,16 @@ def equivalent_frames(
     check_column_order: bool = False,
     cols_case_sensitive: bool = False,
     check_row_order: bool = False,
-    local_data_model=None,
 ) -> bool:
     """return False if the frames are equivalent (up to column re-ordering and possible row-reordering).
     Ignores indexing. None and nan are considered equivalent in numeric contexts."""
     # leave in extra checks as this is usually used by test code
-    if local_data_model is None:
-        local_data_model = data_algebra.pandas_model.default_data_model
+    local_data_model = data_algebra.data_model.lookup_data_model_for_dataframe(a)
+    assert local_data_model.is_appropriate_data_instance(a)
+    assert local_data_model.is_appropriate_data_instance(b)
+    a = local_data_model.to_pandas(a)
+    b = local_data_model.to_pandas(b)
+    local_data_model = data_algebra.data_model.lookup_data_model_for_dataframe(a)
     assert local_data_model.is_appropriate_data_instance(a)
     assert local_data_model.is_appropriate_data_instance(b)
     if a.shape != b.shape:
@@ -88,13 +104,13 @@ def equivalent_frames(
         return True
     if a.equals(b):
         return True
-    a = a.reset_index(drop=True, inplace=False)  # get rid of indices
-    b = b.reset_index(drop=True, inplace=False)
+    a = local_data_model.clean_copy(a)
+    b = local_data_model.clean_copy(b)
     if not cols_case_sensitive:
         a.columns = [c.lower() for c in a.columns]
-        a = a.reset_index(drop=True)
+        a = local_data_model.clean_copy(a)
         b.columns = [c.lower() for c in b.columns]
-        b = b.reset_index(drop=True)
+        b = local_data_model.clean_copy(b)
     a_columns = [c for c in a.columns]
     b_columns = [c for c in b.columns]
     if set(a_columns) != set(b_columns):
@@ -105,14 +121,14 @@ def equivalent_frames(
     else:
         # re-order b into a's column order
         b = b[a_columns]
-        b = b.reset_index(drop=True)
+        b = local_data_model.clean_copy(b)
     if a.shape[0] < 1:
         return True
     if not check_row_order:
         a = a.sort_values(by=a_columns)
-        a = a.reset_index(drop=True)
+        a = local_data_model.clean_copy(a)
         b = b.sort_values(by=a_columns)
-        b = b.reset_index(drop=True)
+        b = local_data_model.clean_copy(b)
     for c in a_columns:
         ca = a[c]
         cb = b[c]
@@ -278,13 +294,11 @@ def _run_handle_experiments(
             raise ValueError(f"{db_handle} ops result did not match expect")
 
 
-# noinspection PyShadowingNames
-def check_transform_on_handles(
+def check_transform_on_data_model(
     *,
     ops,
-    data: dict,
+    data: Dict,
     expect,
-    db_handles,
     float_tol: float = 1e-8,
     check_column_order: bool = False,
     cols_case_sensitive: bool = False,
@@ -299,9 +313,8 @@ def check_transform_on_handles(
     Asserts if there are issues
 
     :param ops: data_algebra.data_ops.ViewRepresentation
-    :param data: pd.DataFrame or map of strings to pd.DataFrame
-    :param expect: pd.DataFrame
-    :param db_handles:  list of database handles to use in testing
+    :param data: DataFrame or map of strings to pd.DataFrame
+    :param expect: DataFrame
     :param float_tol: passed to equivalent_frames()
     :param check_column_order: passed to equivalent_frames()
     :param cols_case_sensitive: passed to equivalent_frames()
@@ -312,16 +325,20 @@ def check_transform_on_handles(
     :param empty_produces_empty: logical, if True assume emtpy inputs should produce empty output
     :return: None, assert if there is an issue
     """
-
     assert isinstance(data, dict)
-    orig_data = {k: v.copy() for k, v in data.items()}
+    assert expect is not None
+    if local_data_model is None:
+        local_data_model = data_algebra.data_model.lookup_data_model_for_dataframe(expect)
+    if not local_data_model.is_appropriate_data_instance(expect):
+        raise TypeError("expected expect to be a DataFrame")
+    assert isinstance(ops, ViewRepresentation)
+    orig_data = {k: local_data_model.clean_copy(v) for k, v in data.items()}
     n_tables = len(data)
     assert n_tables > 0
-    if local_data_model is None:
-        local_data_model = data_algebra.pandas_model.default_data_model
-    assert isinstance(ops, ViewRepresentation)
-    if not local_data_model.is_appropriate_data_instance(expect):
-        raise TypeError("expected expect to be a local_data_model.pd.DataFrame")
+    data_model_map = {local_data_model.presentation_model_name: local_data_model.module}
+    def_pd_data_model = data_algebra.data_model.lookup_data_model_for_key("default_Pandas_model")
+    if def_pd_data_model.presentation_model_name not in data_model_map.keys():
+        data_model_map[def_pd_data_model.presentation_model_name] = def_pd_data_model.module
     cols_used = ops.columns_used()
     if len(cols_used) < 1:
         raise ValueError("no tables used")
@@ -333,7 +350,7 @@ def check_transform_on_handles(
     res = ops.eval(data_map=data)
     if not local_data_model.is_appropriate_data_instance(res):
         raise ValueError(
-            "expected res to be local_data_model.pd.DataFrame, got: " + str(type(res))
+            "expected res to be DataFrame, got: " + str(type(res))
         )
     if not equivalent_frames(
         res,
@@ -343,19 +360,21 @@ def check_transform_on_handles(
         cols_case_sensitive=cols_case_sensitive,
         check_row_order=check_row_order,
     ):
-        raise ValueError("Pandas eval result did not match expect")
+        raise ValueError("local data model eval result did not match expect")
     # show inputs didn't change
     for k, v in orig_data.items():
         v2 = data[k]
         assert equivalent_frames(v, v2)
     if check_parse:
-        if not formats_to_self(ops):
+        if not formats_to_self(ops, data_model_map=data_model_map):
             raise ValueError("ops did not round-trip format")
-        ops_2 = re_parse(ops)
+        ops_2 = _re_parse(
+            ops, 
+            data_model_map=data_model_map)
         res_2 = ops_2.eval(data_map=data)
         if not local_data_model.is_appropriate_data_instance(res_2):
             raise ValueError(
-                "(reparse) expected res to be local_data_model.pd.DataFrame, got: "
+                "(reparse) expected res to be DataFrame, got: "
                 + str(type(res_2))
             )
         if not equivalent_frames(
@@ -366,7 +385,7 @@ def check_transform_on_handles(
             cols_case_sensitive=cols_case_sensitive,
             check_row_order=check_row_order,
         ):
-            raise ValueError("(reparse) Pandas eval result did not match expect")
+            raise ValueError("(re-parse) eval result did not match expect")
     if len(data) == 1:
         res_t = ops.transform(list(data.values())[0])
         if not equivalent_frames(
@@ -377,7 +396,7 @@ def check_transform_on_handles(
             cols_case_sensitive=cols_case_sensitive,
             check_row_order=check_row_order,
         ):
-            raise ValueError("Pandas transform result did not match expect")
+            raise ValueError("local data model transform result did not match expect")
         # show inputs didn't change
         for k, v in orig_data.items():
             v2 = data[k]
@@ -385,7 +404,7 @@ def check_transform_on_handles(
     if valid_for_empty:
         # try on empty inputs
         empty_map = {
-            k: v.iloc[range(0), :].reset_index(drop=True, inplace=False)
+            k: local_data_model.clean_copy(v.head(0))
             for k, v in data.items()
         }
         empty_res = ops.eval(empty_map)
@@ -401,17 +420,56 @@ def check_transform_on_handles(
             keys = [k for k in data.keys()]
             partial_maps = [
                 {
-                    keys[0]: data[keys[0]].reset_index(drop=True, inplace=False),
+                    keys[0]: local_data_model.clean_copy(data[keys[0]]),
                     keys[1]: data[keys[1]],
                 },
                 {
                     keys[0]: data[keys[0]],
-                    keys[1]: data[keys[1]].reset_index(drop=True, inplace=False),
+                    keys[1]: local_data_model.clean_copy(data[keys[1]]),
                 },
             ]
             for pm in partial_maps:
                 empty_res_i = ops.eval(pm)
                 assert set(empty_res_i.columns) == set(res.columns)
+
+
+def _check_transform_on_handles(
+    *,
+    ops,
+    data: Dict,
+    expect,
+    db_handles,
+    float_tol: float = 1e-8,
+    check_column_order: bool = False,
+    cols_case_sensitive: bool = False,
+    check_row_order: bool = False,
+    local_data_model=None,
+) -> None:
+    """
+    Test an operator dag produces the expected result, and parses correctly.
+    Asserts if there are issues
+
+    :param ops: data_algebra.data_ops.ViewRepresentation
+    :param data: DataFrame or map of strings to pd.DataFrame
+    :param expect: DataFrame
+    :param db_handles:  list of database handles to use in testing
+    :param float_tol: passed to equivalent_frames()
+    :param check_column_order: passed to equivalent_frames()
+    :param cols_case_sensitive: passed to equivalent_frames()
+    :param check_row_order: passed to equivalent_frames()
+    :param local_data_model: optional alternate evaluation model
+    :return: None, assert if there is an issue
+    """
+
+    assert isinstance(data, dict)
+    assert expect is not None
+    assert isinstance(ops, ViewRepresentation)
+    if local_data_model is None:
+        local_data_model = data_algebra.data_model.lookup_data_model_for_dataframe(expect)
+    n_tables = len(data)
+    assert n_tables > 0
+    if not local_data_model.is_appropriate_data_instance(expect):
+        raise TypeError("expected expect to be a DataFrame")
     # try any db paths
     global global_test_result_cache
     global run_direct_ops_path_tests
@@ -495,13 +553,16 @@ def check_transform(
     cols_case_sensitive: bool = False,
     check_row_order: bool = False,
     check_parse: bool = True,
+    try_on_DBs: bool = True,
     models_to_skip: Optional[Iterable] = None,
     valid_for_empty: bool = True,
     empty_produces_empty: bool = True,
+    local_data_model = None,
+    try_on_Polars: bool = True,
 ) -> None:
     """
     Test an operator dag produces the expected result, and parses correctly.
-    Assert if there are issues.
+    Assert/raise if there are issues.
 
     :param ops: data_algebra.data_ops.ViewRepresentation
     :param data: pd.DataFrame or map of strings to pd.DataFrame
@@ -511,62 +572,94 @@ def check_transform(
     :param cols_case_sensitive: passed to equivalent_frames()
     :param check_row_order: passed to equivalent_frames()
     :param check_parse: if True check expression parses/formats to self
+    :param try_on_DBs: if true, try on databases
     :param models_to_skip: None or set of model names or models to skip testing
     :param valid_for_empty: logical, if True perform tests on empty inputs
-    :param empty_produces_empty: logical, if True assume emtpy inputs should produce empty output
+    :param empty_produces_empty: logical, if True assume empty inputs should produce empty output
+    :param local_data_mode: data model to use
+    :param try_on_Polars: try tests again on Polars
     :return: nothing
     """
-
     # convert single table to dictionary
     if not isinstance(data, dict):
         cols_used = ops.columns_used()
         table_name = [k for k in cols_used.keys()][0]
         data = {table_name: data}
-
-    db_handles = [
-        # non-connected handles, lets us test some of the SQL generation path
-        data_algebra.SQLite.SQLiteModel().db_handle(None),
-        data_algebra.BigQuery.BigQueryModel().db_handle(None),
-        data_algebra.PostgreSQL.PostgreSQLModel().db_handle(None),
-        data_algebra.SparkSQL.SparkSQLModel().db_handle(None),
-        data_algebra.MySQL.MySQLModel().db_handle(None),
-    ]
-
-    test_dbs = get_test_dbs()
-    db_handles = db_handles + test_dbs
-
-    if models_to_skip is not None:
-        models_to_skip = {str(m) for m in models_to_skip}
-        db_handles = [h for h in db_handles if str(h.db_model) not in models_to_skip]
-
-    caught: Optional[Any] = None
-    try:
-        check_transform_on_handles(
+    assert isinstance(try_on_DBs, bool)
+    assert isinstance(try_on_Polars, bool)
+    assert expect is not None
+    if local_data_model is None:
+        local_data_model = data_algebra.data_model.lookup_data_model_for_dataframe(expect)
+    assert local_data_model.is_appropriate_data_instance(expect)
+    check_transform_on_data_model(
+        ops=ops,
+        data=data,
+        expect=expect,
+        float_tol=float_tol,
+        check_column_order=check_column_order,
+        cols_case_sensitive=cols_case_sensitive,
+        check_row_order=check_row_order,
+        check_parse=check_parse,
+        valid_for_empty=valid_for_empty,
+        empty_produces_empty=empty_produces_empty,
+        local_data_model=local_data_model,
+    )
+    if try_on_Polars and have_polars:
+        polars_data_model = data_algebra.data_model.lookup_data_model_for_key("default_Polars_model")
+        pl = polars_data_model.module
+        expect_polars = pl.DataFrame(expect)
+        data_polars = {k: pl.DataFrame(d) for k, d in data.items()}
+        check_transform_on_data_model(
             ops=ops,
-            data=data,
-            expect=expect,
+            data=data_polars,
+            expect=expect_polars,
             float_tol=float_tol,
             check_column_order=check_column_order,
             cols_case_sensitive=cols_case_sensitive,
             check_row_order=check_row_order,
-            check_parse=check_parse,
-            db_handles=db_handles,
+            check_parse=False,
             valid_for_empty=valid_for_empty,
             empty_produces_empty=empty_produces_empty,
+            local_data_model=polars_data_model,
         )
-    except AssertionError as ase:
-        traceback.print_exc()
-        caught = ase
-    except Exception as exc:
-        traceback.print_exc()
-        caught = exc
-
-    for handle in db_handles:
-        # noinspection PyBroadException
+    if try_on_DBs:
+        caught: Optional[Any] = None
+        db_handles = [
+            # non-connected handles, lets us test some of the SQL generation path
+            data_algebra.SQLite.SQLiteModel().db_handle(None),
+            data_algebra.BigQuery.BigQueryModel().db_handle(None),
+            data_algebra.PostgreSQL.PostgreSQLModel().db_handle(None),
+            data_algebra.SparkSQL.SparkSQLModel().db_handle(None),
+            data_algebra.MySQL.MySQLModel().db_handle(None),
+        ]
         try:
-            handle.close()
-        except Exception:
-            pass
-
-    if caught is not None:
-        raise ValueError("testing caught " + str(caught))
+            test_dbs = get_test_dbs()
+            db_handles = db_handles + test_dbs
+            if models_to_skip is not None:
+                models_to_skip = {str(m) for m in models_to_skip}
+                db_handles = [h for h in db_handles if str(h.db_model) not in models_to_skip]
+            _check_transform_on_handles(
+                ops=ops,
+                data=data,
+                expect=expect,
+                float_tol=float_tol,
+                check_column_order=check_column_order,
+                cols_case_sensitive=cols_case_sensitive,
+                check_row_order=check_row_order,
+                db_handles=db_handles,
+                local_data_model=local_data_model,
+            )
+        except AssertionError as ase:
+            traceback.print_exc()
+            caught = ase
+        except Exception as exc:
+            traceback.print_exc()
+            caught = exc
+        for handle in db_handles:
+            # noinspection PyBroadException
+            try:
+                handle.close()
+            except Exception:
+                pass
+        if caught is not None:
+            raise caught
