@@ -8,7 +8,7 @@ from typing import Dict, Iterable, Optional, Tuple
 
 import numpy
 
-import data_algebra.pandas_model
+import data_algebra.data_model
 
 
 def pandas_to_example_str(obj, *, local_data_model=None) -> str:
@@ -20,11 +20,11 @@ def pandas_to_example_str(obj, *, local_data_model=None) -> str:
     :return: Python source code representation of obj.
     """
     if local_data_model is None:
-        local_data_model = data_algebra.pandas_model.default_data_model
+        local_data_model = data_algebra.data_model.lookup_data_model_for_dataframe(obj)
     pd_module_name = local_data_model.presentation_model_name
     if not local_data_model.is_appropriate_data_instance(obj):
         raise TypeError("Expect obj to be local_data_model.pd.DataFrame")
-    obj = obj.reset_index(drop=True, inplace=False)
+    obj = local_data_model.clean_copy(obj)
     nrow = obj.shape[0]
     pandas_string = pd_module_name + ".DataFrame({"
     for k in obj.columns:
@@ -38,31 +38,6 @@ def pandas_to_example_str(obj, *, local_data_model=None) -> str:
     return pandas_string
 
 
-def table_is_keyed_by_columns(table, column_names: Iterable[str]) -> bool:
-    """
-    Check if a table is keyed by a given list of column names.
-
-    :param table: pandas DataFrame
-    :param column_names: list of column names
-    :return: True if rows are uniquely keyed by values in named columns
-    """
-    # check for ill-condition
-    if isinstance(column_names, str):
-        column_names = [column_names]
-    else:
-        column_names = [c for c in column_names]
-    missing_columns = set(column_names) - set([c for c in table.columns])
-    if len(missing_columns) > 0:
-        raise KeyError("missing columns: " + str(missing_columns))
-    # get rid of some corner cases
-    if table.shape[0] < 2:
-        return True
-    if len(column_names) < 1:
-        return False
-    counts = table.groupby(column_names).size()
-    return max(counts) <= 1
-
-
 # noinspection PyBroadException
 def _mk_type_conversion_table() -> Dict[type, type]:
     """
@@ -162,12 +137,12 @@ def guess_carried_scalar_type(col) -> type:
     }:
         return ct
     # look at a list or Series
-    if isinstance(col, data_algebra.pandas_model.default_data_model.pd.core.series.Series):
+    if isinstance(col, data_algebra.data_model.default_data_model().pd.core.series.Series):
         col = col.values
     if len(col) < 1:
         return type(None)
     good_idx = numpy.where(
-        numpy.logical_not(data_algebra.pandas_model.default_data_model.pd.isna(col))
+        numpy.logical_not(data_algebra.data_model.default_data_model().pd.isna(col))
     )[0]
     test_idx = 0
     if len(good_idx) > 0:
@@ -201,7 +176,7 @@ def guess_column_types(
         if (
             (gt is None)
             or (not isinstance(gt, type))
-            or gt == data_algebra.pandas_model.default_data_model.pd.core.series.Series
+            or gt == data_algebra.data_model.default_data_model().pd.core.series.Series
         ):
             # pandas.concat() poisons types with Series, don't allow that
             return dict()
